/*======================================================================

    daqcard700.c v1.04    (known internally @JHCS as daq700pcs.c)

    9-30-1999

    A daqcard700 PCMCIA real-time client driver

    for Real Time Linux kernels only, at present.
    see http://rtlinux.cs.nmt.edu/~rtlinux    


    The contents of this file are subject to the Mozilla Public
    License Version 1.0 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a copy of
    the License at http://www.mozilla.org/MPL/

    Software distributed under the License is distributed on an "AS
    IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
    implied. See the License for the specific language governing
    rights and limitations under the License.

    v1.00 Originally written by Oleg Subbotin as "daqcard700.c" (oleg@cs.nmt.edu)
    v1.01 Revised by Steve Rosenbluth, JHCS 10-18-98 (stever@silicon.net)
    v1.02 Revised by Steve Rosenbluth, JHCS 3-17-99
    v1.03 Revised by Steve Rosenbluth, JHCS 8-30-99
    v1.04 Revised by Steve Rosenbluth, JHCS 9-30-99

    Revision Notes v1.01 (by Steve): driver is now runtime configurable by writing a 
    config struct to an "incoming" rt-fifo. struct and necessary constants are 
    defined in daqcard700-config.h. configurable options include: burst-mode acq, 
    variable start-conversion timer period, variable banks (num samples per chan),
    choice of voltage range, result number-system choice, option to not use calib 
    constants, acquisition can be stopped and started by user, etc...
    Driver still operates as 1.0 did if you just load it and do a 'cardctl reset'
    A sample configuration app 'daq-config' is included in source form.
    Lots of commenting was done.
    Many identifiers have been re-named, constants capitalized, variables introduced,
    etc (you probably can't 'diff' 1.0 and 1.01).
    D. Hinds' redefine of the X macro has been adopted because we couldn't compile 
    without it on 2.0.35 with modversions.
    Bug fix: input_mode is ANDed when used (to clear unwanted bits).
    Bug fix: gain_error math parenthesis have been fixed.
    Some other logical bugs are fixed..., maybe 'diff' will tell you what they are.
    Many things are left undone - check the source, and use at your own risk. 

    Revision Notes v1.02 (by Steve):
    * Drivers 1.01 and 1.02 successfuly compile with kernel 2.0.35 and pcmcia-cs-3.0.7
    * Now supports digital output: Send a byte to daqcard driver through the 
    config fifo and the 5 MSB's of it will appear on the digital output lines of the
    card when the next interrupt occurs (3 LSB's are reserved for multiplexing analog
    and digital inputs). see the daqcard700-config.h file for config struct format,
    and try "daq-config" sample app.
    * Now supports digital input: Any change in the logic levels on the digital input
    byte of the card is written to a digital input result rt-fifo on the next 
    interrupt. A new config field specifies one or two banks of digital input.
    Try the "daq-config" sample app and do "cat /dev/rtf2" .
    * Now supports two different number systems for adc results
    * Known bug: on some systems, the first conversion is sometimes missing from a 
    multi-channel, multi-bank sample burst.
    * Known bug: 'cardctl reset' might permanently disable burst mode ...    

    Revision Notes v1.03 (by Steve):
    * Digital output access is changed. Now all 8 bits are available if not using
    multiple banks of ADC or DIO. Otherwise, 6 MSBs available for digital output.
    * Change in config structure and daqcard700_dev_t variable names, new vars too.
    * Major changes were made to daqcard700_interrupt() function. It now addresses up
    to 4 banks of analog and up to 4 banks of digital inputs. "Global" vars were taken
    out of handler and given to device-specific structure for support of multiple
    daqcards. daqcard_config() now inits some of those vars.
    * Digital input banks use the same 2 digital out LSBs to switch an external mux.
    * Fixed data buffer dereferencing bug and a non-static variable bug in handler.
    * Added & revised many comments in irq handler.
    * Added a work-around for the "banks bug", where chan15 comes back as chan14, etc
    in multi-bank mode. Users can set this with insmod if they need it:
    insmod need_banks_delay=1, or put in /etc/pcmcia/config.opts: 
    module "daqcard700" opts "need_banks_delay=1"
    * Removed redundant channel-scan disabling when we disable interrupts in handler
    * Hot-swapping appears to work OK now, with k2.0.35 and pcmcia-cs 3.0.12

    Revision Notes v1.04 (by Steve):
    * Ported to kernel 2.2.12 , RTlinuxbeta15, card services 3.1.0, gcc 2.95.1
    All previous revisions were only tested on k2.0.35 w/ pcmcia 3.0.12
    * Check and report errors on all rtlinux API calls in init_module() and
    daqcard700_config()
    * Print errors when kmallocs fail in daqcard700_attach()
    * Use CardServices() return constants in daqcard700_attach()
    * daqcard700_detach(): changed error message, still not done w/ locking issue.
    * Had to add "daq" class file to /etc/pcmcia for cs to succeed in loading this 
    driver
    * fixed messed up _release() stuff in _detach(). now _detach() sets stale_link
    flag and returns, allowing cleanup_module() to call it later.
    * added (crude) error-checking for data errors
    * irq handler: fixed missing reentry_stat reset when devlist == NULL
    * Finally fixed the "banks bug" ! This is a major change: we no longer disable
    irqs on adc chipset.
    Now, bursts are started/stopped by en/disabling the timer chip which paces
    conversions.
    * Fixed bug where switching between continuous and burst mode caused
    conversions to be out of order. 'Stop' resets chan_cnt and mux now.
    * Reimplemented 'burst_done" flag in dev_info and irq handler.
    * Probably fixed the "channel shift" bug where conversions got out of order.
    Each irq handler now resets ADC channel_scan and clears adc/fifo at end of a 
    bank, also 'Start' now resets ADC channel_scan and clears adc/fifo.
    'Start' now also avoids doing this if in continuous mode.
    
    Copyright(C) 1998,1999 Jim Henson's Creature Shop.
    Except original daqcard700_interrupt() function, v1.00 - Copyright(C) 1998, Oleg Subbotin.

======================================================================*/
/*  epson dot-matrix printer condense code */

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/cons.h>
#include <linux/errno.h>
#include <asm/io.h>
#include <asm/system.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ds.h>

/* Real-time include files */
/*#include </usr/src/rtl/include/rtl_sched.h>*/
#include <rtlinux/rtl_sched.h>
/*#include </usr/src/rtl/include/rtl_fifo.h>*/
#include <rtlinux/rtl_fifo.h>
/*#include <asm/rt_irq.h>*/
#include <rtlinux/arch/rt_irq.h>

/*#include "daq700pcs.h"*/
/*#include "daq700pcs-config.h" */                                                 /* defs for fifo control of driver */
#include "daqcard700.h"
#include "daqcard700-config.h"

/*
 * All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
 * you do not define PCMCIA_DEBUG at all, all the debug code will be
 * left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
 * be present but disabled -- but it can then be enabled for specific
 * modules at load time with a 'pc_debug=#' option to insmod.
 */
#define PCMCIA_DEBUG 1
#define SCOPEING 0                                                          /* set this to 1 to see irq duration on an oscilloscope */
#define VERBOSE_MESSAGING 0                                               /* set this to be verbose receiving config messages */

#ifdef SCOPEING
    #define LP_PORT 0x378
#endif

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_INFO args);
#else
#define DEBUG(n, args...)
#endif

#define CS_CHECK(fn, args...) \
while ((last_ret=CardServices(last_fn=(fn),args))!=0) goto cs_failed

#define CFG_CHECK(fn, args...) \
if (CardServices(fn, args) != 0) goto next_entry

/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* The old way: bit map of interrupts to choose from */
/* This means pick from 15, 14, 12, 11, 10, 9, 7 */
static u_int irq_mask = 0xde80;
/* Newer, simpler way of listing specific interrupts */
static int irq_list[4] =
{-1};
/* set this with: insmod need_banks_delay=1, or put in /etc/pcmcia/config.opts: module "daqcard700" opts "need_banks_delay=1" */
/*static int need_banks_delay = FALSE;*/                                         /* set this to 1 to try to cure banks bug */

/* DaqCard specific pseudo-constants */
static int INSTALL_HANDLER = 1;
static int TIMER_PERIOD = 50;                                                 /* conver rate in microsecs (50=20kHz)(from 1Mhz clock) */
static int MAX_CHAN_SCAN = 0x0f;                                              /* num physical channels we can scan for samples */
static int INPUT_MODE = 0x0;                                                  /* bitmask for cmd reg 3. deflt: 10Vrange, single-ended */

MODULE_PARM(irq_mask, "i");                                                    /* these make values changeable from insmod, modprobe */
MODULE_PARM(irq_list, "1-6i");
MODULE_PARM(INSTALL_HANDLER, "i");
MODULE_PARM(TIMER_PERIOD, "i");
MODULE_PARM(MAX_CHAN_SCAN, "i");
MODULE_PARM(INPUT_MODE, "i");
/*MODULE_PARM(need_banks_delay, "i");*/

int first_start_acq=TRUE;                                           /* to reset the daqcard once per config */
/*int fifocount_max=0;*/    /* for testing only ?!? */


/*====================================================================*/
/***********function prototypes*************/
/*
 * The event() function is this driver's Card Services event handler.
 * It will be called by Card Services when an appropriate card status
 * event is received.  The config() and release() entry points are
 * used to configure or release a socket, in response to card
 * insertion and ejection events.  They are invoked from the dummy
 * event handler. 
 */

static int daqcard700_event(event_t event, int priority,
                            event_callback_args_t * args);
static void daqcard700_config(dev_link_t * link);
static void daqcard700_release(u_long arg);
int convert(int value, struct daq_error *error, int numsys);       /* take raw adc conversion val, correct for error and number system */


/*
 * The attach() and detach() entry points are used to create and destroy
 * "instances" of the driver, where each instance represents everything
 * needed to manage one actual PCMCIA card.
 */

static dev_link_t *daqcard700_attach(void);                                   /* dev_link_t defined in card services ds.h */
static void daqcard700_detach(dev_link_t *);

/*
 * functions that actually talk to the device. 
*/

static void daqcard700_interrupt(void);                                       /* invoked when daqcard generates a hardware interrupt  */
int Ctl_fifo_handler(unsigned int fifo);                                      /* handler invoked in rt task when fifo accessed */


/*************global variables**********************/

/*
 * The dev_info variable is the "key" that is used to match up this
 * device driver with appropriate cards, through the card configuration
 * database.
 */

static dev_info_t dev_info = "daqcard700";

/*
 * A linked list of "instances" of the daqcard700 device.  Each actual
 * PCMCIA card corresponds to one device instance, and is described
 * by one dev_link_t structure (defined in ds.h).
 * 
 * We can have only one (for now) daqcard700 device.
 * 
 * dev_link_t is defined in /pcmcia/ds.h, it contains a pointer to our own
 * struct describing our card "daqcard700_dev_t" as ->priv member
 */

static dev_link_t *dev_list = NULL;

/*
 * A dev_link_t structure has fields for most things that are needed
 * to keep track of a socket, but there will usually be some device
 * specific information that also needs to be kept track of.  The
 * 'priv' pointer in a dev_link_t structure can be used to point to
 * a device-specific private data structure, like this.
 */

/* Export some functions and variables */
/* User API functions prototypes */

#if (LINUX_VERSION_CODE <= VERSION(2,1,17))
#undef CONFIG_MODVERSIONS
static struct symbol_table daqcard700_symtab =
{
#include <linux/symtab_begin.h>
#if (LINUX_VERSION_CODE >= VERSION(1,3,0))
#undef X
#define X(sym) { (void *)&sym, SYMBOL_NAME_STR(sym) }
#endif
    X(daqcard700_reset),
    X(daqcard700_write_register),
    X(daqcard700_read_register),
    X(daqcard700_read_fifo),
    X(daqcard700_stop_timer),
#include <linux/symtab_end.h>
};
#else
#define register_symtab(n)
EXPORT_SYMBOL(daqcard700_reset);
EXPORT_SYMBOL(daqcard700_write_register);
EXPORT_SYMBOL(daqcard700_read_register);
EXPORT_SYMBOL(daqcard700_read_fifo);
EXPORT_SYMBOL(daqcard700_stop_timer);
#endif

static char *version =
"daqcard700.c v1.04 9-30-1999 (by Oleg Subbotin & Steve Rosenbluth)";

/*================================= functions ===============================*/


#define daqcard700_clear_fifo(base) \
       outb(0x00,  base+DAQ_AD_CLR_REG); \
       while (inb(base+DAQ_STAT_REG_1) & ConversionInProgress); \
       inw(base+DAQ_AD_FIFO_REG);


/******************************************/
static void cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err =
    {func, ret};
    CardServices(ReportError, handle, &err);
}

/*======================================================================

    daqcard700_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services and called by card services.

    The dev_link structure is initialized, but we don't actually
    configure the card at this point -- we wait until we receive a
    card insertion event.
    
======================================================================*/
/*********************************/
static dev_link_t *daqcard700_attach(void)
{
    client_reg_t client_reg;
    dev_link_t *link;
    daqcard700_dev_t *info;
    int ret, i;

    DEBUG(0, "daqcard700_attach()\n");

    /* Initialize the dev_link_t structure */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    if (!link)
    {
        printk("bad kmalloc of link struct!\n"); /* for testing only ?!? */
        return NULL;
    }

    memset(link, 0, sizeof(struct dev_link_t));
    link->release.function = &daqcard700_release;
    link->release.data = (u_long) link;

    /* Interrupt setup */
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
    link->irq.IRQInfo1 = IRQ_PULSE_ID | IRQ_INFO2_VALID;

    if (irq_list[0] == -1)
        link->irq.IRQInfo2 = irq_mask;
    else
        for (i = 0; i < 4; i++)
            link->irq.IRQInfo2 |= 1 << irq_list[i];

    /* IO setup */
    link->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
    link->io.NumPorts1 = 0x20;
    link->io.IOAddrLines = 5;

    /* 
     * General socket configuration defaults can go here.  In this client, we assume very little, and rely on the CIS for almost
     * everything.  In most clients, many details (i.e., number, sizes, and attributes of IO windows) are fixed by the nature of
     * the device, and can be hard-wired here. */

    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;

    /* Allocate space for private device-specific data */
    info = kmalloc(sizeof(daqcard700_dev_t), GFP_KERNEL);
    if (!info)
    {
        kfree_s(link, sizeof(struct dev_link_t));
        printk("bad kmalloc of info struct!\n"); /* for testing only ?!? */
        return NULL;
    }

    memset(info, 0, sizeof(daqcard700_dev_t));
    link->priv = info;
    link->next = dev_list;
    dev_list = link;                                                           /* first node of our list of card instances */

    /* Register with Card Services */

    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =
        CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
        CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
        CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &daqcard700_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;

    ret = CardServices(RegisterClient, &link->handle, &client_reg);

    if (ret != CS_SUCCESS) /* was (ret != 0) */
    {
        cs_error(link->handle, RegisterClient, ret);
        daqcard700_detach(link);
        DEBUG(0, "daqcard700_attach() failed...\n");
        return NULL;
    }

    DEBUG(0, "daqcard700_attach() ok...\n");
    return link;
}                                                                              /* daqcard700_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/
/*************************************/
static void daqcard700_detach(dev_link_t * link)
{
    dev_link_t **linkp;

    DEBUG(0, "daqcard700_detach(0x%p)\n", link);

    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)                   /* run through llist of card instances */
        if (*linkp == link)
            break;

    if (*linkp == NULL)
        return;

    /*
       If the device is currently configured and active, we won't
       actually delete it yet.  Instead, it is marked so that when
       the release() function is called, that will trigger a proper
       detach().
    */
    if (link->state & DEV_CONFIG)
    {
        #if (PCMCIA_DEBUG > 1)
        printk(KERN_DEBUG "dacard700: detach postponed, '%s' "
               "still locked\n", link->dev->dev_name);
        #endif
        /*daqcard700_release((u_long) link); was this */
        link->state |= DEV_STALE_LINK;          /* new for 1.04 */
        return;                                 /* new for 1.04 */
    }

    /* Break the link with Card Services */
    if (link->handle)
        CardServices(DeregisterClient, link->handle);

    /* Unlink device structure, free pieces */
    *linkp = link->next;
    if (link->priv)
        kfree_s(link->priv, sizeof(daqcard700_dev_t));

    kfree_s(link, sizeof(struct dev_link_t));
}                                                                              /* daqcard700_detach */

/* called from daqcard700_config */
/********************************************/
void DAQ_Get_Calibration_Result(tuple_t * tuple, daqcard700_dev_t * info)
{
    signed char *data = tuple->TupleData;                               /* should be signed for portability */
    int index = (unsigned int) data[0];

    /*DEBUG(0, "Filling calibration Table with code: %d value %d %d\n",
          (int) data[0], (int) data[1], (int) data[2]);*/

    index = (index == 0xff) ? 8 : index;

    info->calib[index].valid = 1;
    info->calib[index].offset_error = (int) (data[1] / 2);                  /* offset error stored on card is multiplied by 2 */
    info->calib[index].gain_error = (int) (data[2]);
}


/*======================================================================

    daqcard700_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    device available to the system.  It installs the real time interrupt handler.
    Called when a card insertion event occurs(when machine boots up w/ card too). 
    
======================================================================*/
/*******************************************/
static void daqcard700_config(dev_link_t * link)
{
    client_handle_t handle;
    tuple_t tuple;
    cisparse_t parse;
    daqcard700_dev_t *info;
    int last_fn, last_ret, i, ret;
    u_char buf[64];

    handle = link->handle;
    info = (daqcard700_dev_t *) link->priv;

    DEBUG(0, "daqcard700_config(0x%p)\n", link);

    /* 
     * This reads the card's CONFIG tuple to find its configuration registers. */
    tuple.DesiredTuple = CISTPL_CONFIG;
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = sizeof(buf);
    tuple.TupleOffset = 0;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];


    /* Configure card */
    /* Getting calibration results from CIS tuples */

    link->state |= DEV_CONFIG;

    tuple.DesiredTuple = CISTPL_ADC_CALIBRATION;

    if (CardServices(GetFirstTuple, handle, &tuple) == CS_SUCCESS)
    {
        CardServices(GetTupleData, handle, &tuple);
        DAQ_Get_Calibration_Result(&tuple, info);
        while (CardServices(GetNextTuple, handle, &tuple)
               != CS_NO_MORE_ITEMS)
        {
            CardServices(GetTupleData, handle, &tuple);
            DAQ_Get_Calibration_Result(&tuple, info);
        }
    }

    DEBUG(0, "Reading config table\n");
    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    while (1)
    {
        cistpl_cftable_entry_t dflt =
        {0};
        cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);
        CFG_CHECK(GetTupleData, handle, &tuple);
        CFG_CHECK(ParseTuple, handle, &tuple, &parse);

        if (cfg->index == 0)
            goto next_entry;
        link->conf.ConfigIndex = cfg->index;

        /* Does this card need audio output? */
        if (cfg->flags & CISTPL_CFTABLE_AUDIO)
        {
            link->conf.Attributes |= CONF_ENABLE_SPKR;
            link->conf.Status = CCSR_AUDIO_ENA;
        }

        break;

      next_entry:
        if (cfg->flags & CISTPL_CFTABLE_DEFAULT)
            dflt = *cfg;
        CS_CHECK(GetNextTuple, handle, &tuple);
    }

    /* This reserves IO space but doesn't actually enable it */
    CFG_CHECK(RequestIO, link->handle, &link->io);
    /* 
     * Allocate an interrupt line.  Note that this does not assign a handler to the interrupt, unless the 'Handler' member of the irq
     * structure is initialized. */
    if (link->conf.Attributes & CONF_ENABLE_IRQ)
        CS_CHECK(RequestIRQ, link->handle, &link->irq);

    /* 
     * This actually configures the PCMCIA socket -- setting up the I/O windows and the interrupt mapping, and putting the card and host
     * interface into "Memory and IO" mode. */
    CS_CHECK(RequestConfiguration, link->handle, &link->conf);

    info->irq = link->irq.AssignedIRQ;
    info->base_address = link->io.BasePort1;

    info->input_mode = INPUT_MODE;                                             /* store our desired daq input mode in global struct */
    info->timer_period = TIMER_PERIOD;                                         /* put our timer constant into global struct */
    info->channel_scan = MAX_CHAN_SCAN;                                        /* put our desired num chans in global struct */
    info->num_adc_banks = 1;                                                       /* default: don't use multiple banks */
    info->burst_acq = FALSE;                                                   /* by default do continuous acquisition*/
    info->result_numsys = BTC_NUMSYS;                                        /* default: use binary twos complement conversion results */
    for (i=0; i < 8; i++)                                                      /* for each voltage mode, set default */
        if ((i != 1) && (i != 5))                                              /* these 2 don't exist */
            (info->calib[i]).correcting = TRUE;                                /* we're using hardware calibration constants on results*/
    info->dio_output_val = 0;                                                  /* start with a cleared digital output byte */
    info->num_dig_in_banks = 1;                                                /* accquire only one bank of digital inputs by default */
    /*info->old_digital_input_word = -1;*/                                         /* invalid sample value: so we get a first sample */
    info->multiplexing = FALSE;                                                /* don't assume we're getting multiple banks of samples */
    info->send_dig_ins = TRUE;                                                 /* force driver to send dig input results when loaded */
    info->chan_cnt = 0;
    info->mux = 0;
    info->sample_bank[0] = DAQ_MAGIC;
    /*info->reenable_irq = TRUE;*/
    /*info->sample_request_id = -1;*/                                              /* this makes first sample request zero */ 
    info->burst_done = FALSE;

    if (INSTALL_HANDLER)                                                       /* if we're using an interrupt handler */
    {

        ret = request_RTirq(info->irq, daqcard700_interrupt);                    /* install our handler with the real time executive */
        if (ret)                                                            /* nonzero return means problems */
        {
            printk(KERN_WARNING "ERROR installing RT interrupt handler: !\n");
            if (ret == -EBUSY)
                printk("RT-Interrupt is already in use.");
            else if (ret == -EINVAL)
                printk("Invalid interrupt number, or null handler.");
        }
    }

    /* 
     * At this point, the dev_node_t structure(s) need to be initialized and arranged in a linked list at link->dev. */

    sprintf(info->node.dev_name, "daq700");
    info->node.major = info->node.minor = 0;
    link->dev = &info->node;
    link->state &= ~DEV_CONFIG_PENDING;

    /* Finally, report what we've done */
    printk(KERN_INFO "%s: index 0x%02x: Vcc %d.%d",
           info->node.dev_name, link->conf.ConfigIndex,
           link->conf.Vcc / 10, link->conf.Vcc % 10);
    if (link->conf.Vpp1)
        printk(", Vpp %d.%d", link->conf.Vpp1 / 10, link->conf.Vpp1 % 10);
    if (link->conf.Attributes & CONF_ENABLE_IRQ)
        printk(", irq %d", link->irq.AssignedIRQ);
    if (link->io.NumPorts1)
        printk(", io 0x%04x-0x%04x\n", link->io.BasePort1,
               link->io.BasePort1 + link->io.NumPorts1 - 1);
    printk(KERN_INFO "%s: timer_period %d, channel_scan %d, input_mode 0x%02x\n",
           info->node.dev_name, info->timer_period, info->channel_scan,
           info->input_mode);

    return;

  cs_failed:
    cs_error(link->handle, last_fn, last_ret);
    daqcard700_release((u_long) link);

}                                                                              /* end daqcard700_config */

/*======================================================================

    After a card is removed, daqcard700_release() will unregister the
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
======================================================================*/
/******************************************/
static void daqcard700_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *) arg;

    DEBUG(0, "daqcard700_release(0x%p)\n", link);
    /*printk("fifocount_max is %d\n",fifocount_max);*/ /* for testing only */

    /* 
     * If the device is currently in use, we won't release until it is actually closed, because until then, we can't be sure that no one
     * will try to access the device or its data structures. */
    if (link->open)
    {
        DEBUG(1, "daqcard700: release postponed, '%s' still open\n",
              link->dev->dev_name);
        link->state |= DEV_STALE_CONFIG;
        return;
    }

    /* Unlink the device chain */
    link->dev = NULL;

    /* Stop Data acquisition */
    outb(0x00, ((daqcard700_dev_t *) link->priv)->base_address + DAQ_CMD_REG_1);
    /* ? set input voltage range to +/-5V . Why ?!? */
    outb(0x02, ((daqcard700_dev_t *) link->priv)->base_address + DAQ_CMD_REG_3);

    if (INSTALL_HANDLER)
    {
        free_RTirq(((daqcard700_dev_t *) link->priv)->irq);                    /* unintall our interrupt handler */
    }

    CardServices(ReleaseConfiguration, link->handle);
    CardServices(ReleaseIO, link->handle, &link->io);
    CardServices(ReleaseIRQ, link->handle, &link->irq);
    link->state &= ~DEV_CONFIG;

    if (link->state & DEV_STALE_LINK)
        daqcard700_detach(link);

}                                                                              /* daqcard700_release */

/* This is API to work with the card - to set different modes,
 * write commands and data 
 */

/* Function is called from daqcard700_interrupt , daqcard700_readfifo */
/* This function takes raw adc conversion val and corrects for error and number system */
/* NOTE: error correction did not work correctly when implemented per the National Instruments Programmer manual. */
/* Results were non-linear over the applied voltage sweep in 5 volt mode. Error correction has thus been made optional. */
/* FIXME: BTC mode doesn't disable errors. */
/********************************************/
int convert(int value, struct daq_error *error, int numsys)                 
{
    if (numsys == BOB_NUMSYS)                                               /* Bipolar Offset Binary. 0 to 4095 is -Vmin to +Vmax */
    {
        value &= 0x0fff;                                                    /* clear out high nibble */
        value ^= INVERT_MSB_12;                                             /* invert MSB of 12 bit result to get BOB from BTC */
        if (error->correcting && (error->valid))                            /* if we have no error correction constants */
        {
            value -= error->offset_error;                                   /* subtract precalibrated hardware offset error */
            value -= 2048;                                                  /* values less than half scale become negative */
            value += (value) * ((float)error->gain_error / 4096.0);       /* subtract the precalibrated hardware gain error from value */
            value += 2048;                                                  /* convert back to BOB */
            if (value > 4095)                                               /* clip result if gain arithmetic went overrange */
                value = 4095;
            else if (value < 0)
                value = 0;
        }
        else
            return value;                                                   /* return the raw conversion value */
    }
    else if (numsys == BTC_NUMSYS)                  /* Binary Twos Complement. F800 to 0000 is -Vmin to 0V, 0000 to 7FF is 0V to +Vmax */
    {
        if (value & 0xf000)
            value = -(~value & 0xfff);                           /* ? forcefully convert it to decimal from Binary Twos Complement ?!? */
        if (!error->valid)                                                      /* if we have no error correction constants */
            return value;
        value -= error->offset_error;
        value += value * (error->gain_error / 4096);
        if (value > 4095)                                               /* clip result if gain arithmetic went overrange */
            value = 4095;
        else if (value < 0)
            value = 0;
    }

    return value;
}

/*******************************************/
unsigned char daqcard700_read_register(unsigned char reg)
{
    daqcard700_dev_t *dev_info;
    ioaddr_t base_address;

    if (dev_list == NULL)                                                      /* if there is a list of instances of the card */
    {
        printk(KERN_WARNING "DaqCard has not been initialized properly!!");
        return 0xff;
    }

    dev_info = (daqcard700_dev_t *) dev_list->priv;
    base_address = dev_info->base_address;

    return inb(base_address + reg);
}

/********************************************/
int daqcard700_write_register(unsigned char command, unsigned char reg)
{
    daqcard700_dev_t *dev_info;
    ioaddr_t base_address;

    if (dev_list == NULL)                                                      /* if there is a list of instances of the card */
    {
        printk(KERN_WARNING "DaqCard has not been initialized properly!!");
        return 1;
    }

    dev_info = (daqcard700_dev_t *) dev_list->priv;
    base_address = dev_info->base_address;

    if (reg == DAQ_AD_CLR_REG)
    {
        daqcard700_clear_fifo(base_address);                                   /* reset adc and clear hardware fifo */
    }
    else
        outb(command, base_address + reg);

    return 0;
}

/*******************************************/
int daqcard700_read_fifo(int *buffer, int count, int clear)
{
    daqcard700_dev_t *dev_info;
    ioaddr_t base_address;
    volatile char status = 0;
    volatile int temp = 0;

    int i = 0;

    if (dev_list == NULL)                                                      /* if there is a list of instances of the card */
    {
        printk(KERN_WARNING "DaqCard has not been initialized properly!!");
        return 0;
    }

    dev_info = (daqcard700_dev_t *) dev_list->priv;
    base_address = dev_info->base_address;

    status = inb(base_address + DAQ_STAT_REG_1);

    for (i = 0; i < count; i++)
    {
        if (!(status & DataError) && (status & DATA_AVAIL))
        {
            temp = inw(base_address + DAQ_AD_FIFO_REG);                        /* get conversion result out of hdware fifo */
            buffer[i] = convert(temp, &dev_info->calib[dev_info->input_mode & 0x03],dev_info->result_numsys);
            status = inb(base_address + DAQ_STAT_REG_1);
        }
        else
        {
            if (!(status & DATA_AVAIL))
                return i + 2000;
            /* else return error code */
            status = inb(base_address + DAQ_STAT2);
            daqcard700_clear_fifo(base_address);
            return -(status & 0x03);
        }
    }

    if (clear)
    {
        daqcard700_clear_fifo(base_address);
    }

    return i;
}

/******************************/
void daqcard700_reset(void)
{
    daqcard700_dev_t *dev_info;
    ioaddr_t base_address;

    /* printk(KERN_WARNING "DaqCard: reset\n"); *//* message to raw console */

    if (dev_list == NULL)                                                      /* if there is a list of instances of the card */
    {
        printk(KERN_WARNING "DaqCard has not been initialized properly!!");
        return;
    }

    dev_info = (daqcard700_dev_t *) dev_list->priv;                            /* get addr of our daq struct */
    base_address = dev_info->base_address;                                     /* get io addr of daq card */

    /* Clear timer interrupts and set timer mode */
    outb(0x00, base_address + TIMR_CLR_REG_ADDR);
    /* timer mode: counter0 (first), r/w endian: LSB then MSB, rate generator mode. this stops timer */
    outb(0x34, base_address + CNTR_MODE_REG);

    /* set to channel scanning mode: write w/ MSB set, then reset. enable interrupt generation (by FIFO) when conversion results avail.
     * set chan scanning to MAX_CHAN_SCAN (16) samples.  external interrupts are disabled. */
    outb(ScanEnable | FifoIntEnable | dev_info->channel_scan, base_address + DAQ_CMD_REG_1);
    outb(FifoIntEnable | dev_info->channel_scan, base_address + DAQ_CMD_REG_1);
    outb(0x00, base_address + DAQ_CMD_REG_2);                                  /* enable data acquisition */
    /* FIXME: put OR masks in here */
    /* set FIFO to generate interrupt when >=1 sample available. CLK1 signal comes from IO connector. set to single-ended input mode.
     * default analog input range is +/- 10V. */
    outb(dev_info->input_mode, base_address + DAQ_CMD_REG_3);
    /* load data acq timer 0 of 8254 chip w/ LSByte of our period. (this stops timer) */
    outb(dev_info->timer_period & 0xff, base_address + DAQ_CNTR_0_REG);
    /* clear hardware FIFO, put latest conversion into fifo, clear stat reg error bits */
    daqcard700_clear_fifo(base_address);                                       /* do this while timer is didsabled */
    /* write high byte of our period (zero) into timer ? */
    outb((dev_info->timer_period >> 8) & 0xff, base_address + DAQ_CNTR_0_REG);
}

/******************************/
void daqcard700_stop_timer()
{
    daqcard700_dev_t *dev_info;
    ioaddr_t base_address;

    if (dev_list == NULL)                                                      /* if there is a list of instances of the card */
    {
        printk(KERN_WARNING "DaqCard has not been initialized properly!!");
        return;
    }

    dev_info = (daqcard700_dev_t *) dev_list->priv;                            /* our daq700 struct */
    base_address = dev_info->base_address;                                     /* io addr of the card */

    /* Clear timer interrupts */
    outb(0x00, base_address + TIMR_CLR_REG_ADDR);                              /* clear timer IRQ */
    /* timer mode: counter0 (first),  r/w endian: LSB then MSB, rate generator mode */
    outb(0x34, base_address + CNTR_MODE_REG);
    /* FIXME: should we write lsb of timer value here ?!? */
}


/* This is the handler called when the daqcard generates an interrupt, as when conversion results are in. */
/* Theory of operation: 
   daqcard700_fifo_handler() starts ADC conversions by starting the card's conversion-pacing timer.
   This handler gets called on each conversion of a channel scan (bank), then releases the processor briefly until the next 
   conversion comes in. This wastes a lot of processor time, but it works, for now.
   If we are scanning all 16 channels, one "bank" of analog samples gets written to the rtfifo every 16 interrupts.
   Digital inputs are written to an rtfifo when they change.
   Note: muxes and digital ouputs only change state when there is an interrupt !! */
/*********************************/
static void daqcard700_interrupt()
{
    daqcard700_dev_t *dev_info;
    ioaddr_t base_address;
    static int reentry_stat = 0;
    volatile int temp = 0;
    volatile char status = 0;

    if (reentry_stat)
    {
        conpr("Re-entering IRQ ! abort.\n");
        return;
    }
    reentry_stat = 1;                                                          /* flag we are now in irq handler */

    if (dev_list == NULL)                                                      /* if there is a list of instances of the card */
    {
        printk(KERN_WARNING "DaqCard has not been initialized properly!!");
        reentry_stat = 0;                                                      /* flag done w/ handler */
        return;
    }

#if SCOPEING
    outb(1,LP_PORT);                                               /* send the byte out the port */
#endif

    dev_info = (daqcard700_dev_t *) dev_list->priv;                            /* get our struct describing card */

    if (!dev_info->start)
    {                                                                          /* if the device is not started, (available) */
        /*printk("daq700: Throwing away interrupt, device not started !\n");*/
        reentry_stat = 0;                                                      /* flag done w/ handler */
        return;                                                                /* abort handler */
    }

    base_address = dev_info->base_address;                                 /* get base address of daq card - used throughout this func */

    /* while there are conversions in fifo and havent yet got one set of chans */
    /* Note: this apparently dangerous,time-wasting loop has been tested and doesn't iterate more than 2 times */
    status = inb(base_address + DAQ_STAT_REG_1);                               /* get conversion status bits */

    while ((status & DATA_AVAIL) && (dev_info->chan_cnt <= dev_info->channel_scan))
    {
        temp = inw(base_address + DAQ_AD_FIFO_REG);                            /* get raw conversion data out of hardware fifo */
        /* store conversion value, increment channel counter, adjust for device specific calibration constants */
        dev_info->sample_bank[2 + dev_info->chan_cnt++] = convert(temp,
                                              &dev_info->calib[dev_info->input_mode & 0x03],dev_info->result_numsys);
        if (status & 0x02)                                                  /* this bit signals a data error */
        {
            DEBUG(0,"Error in stat reg 1!\n");            /* FIXME: check which error */
            daqcard700_clear_fifo(base_address);                               /* reset adc and clear hardware fifo */
        }
        status = inb(base_address + DAQ_STAT_REG_1);                           /* see if conversions available */
    }

    if (dev_info->chan_cnt >= (dev_info->channel_scan + 1))                 /* got a full bank of samples */
    {                                                                       /* tested: we don't normally get too many conversions */
        dev_info->sample_bank[1] = dev_info->mux;                           /* stamp the packet: which bank of samples this is */

        dev_info->digital_input_word[dev_info->mux] = (unsigned char) inb(base_address+DAQ_DIGITAL_IN); /* get curr byte of dig input */
        if (dev_info->digital_input_word[dev_info->mux] != dev_info->old_digital_input_word[dev_info->mux]) /* if inputs have changed */
        {
            dev_info->old_digital_input_word[dev_info->mux] = dev_info->digital_input_word[dev_info->mux]; /* store for future compare */
            dev_info->send_dig_ins = TRUE;                                                    /* raise flag that we must send new data */
        }

        /* we enter this block when we finish adc sampling of all banks (burst is done) */
        if (dev_info->mux >= (dev_info->num_adc_banks - 1))                            /* done with all banks */
        {
            /* done with adc banks but havent yet got all the digital inputs */
            if (dev_info->num_dig_in_banks > dev_info->num_adc_banks)
            {
                for (; dev_info->mux < dev_info->num_dig_in_banks; dev_info->mux++)
                {
                    /* get current byte of digital input */
                    dev_info->digital_input_word[dev_info->mux] = (unsigned char) inb(base_address+DAQ_DIGITAL_IN);
                    /* if inputs have changed */
                    if (dev_info->digital_input_word[dev_info->mux] != dev_info->old_digital_input_word[dev_info->mux]) 
                    {
                            /* store value for future comaparison */
                            dev_info->old_digital_input_word[dev_info->mux] = dev_info->digital_input_word[dev_info->mux];
                            dev_info->send_dig_ins = TRUE;                                    /* raise flag that we must send new data */
                    }
                }
            }
            if (dev_info->send_dig_ins)                                                       /* must send all digital input results */
            {
                /* write all digital input bytes to rt-fifo 2 */
                rtf_put(DIG_RESULT_FIFO_ID, dev_info->digital_input_word, dev_info->num_dig_in_banks*sizeof(unsigned char)); 
                dev_info->send_dig_ins = FALSE;
            }

            dev_info->mux = 0;                                                /* reset for first bank next time around */

            if (dev_info->burst_acq)                                          /* if we're in burst mode */
            {
                dev_info->start = 0;                                               /* mark device as "stopped", not available */
                dev_info->burst_done = TRUE;                                   /* flag that we don't want to generate irqs anymore*/
            }
        }
        else                                                                /* haven't yet got full burst of channels */
        {
            dev_info->mux += 1;                                              /* increment to next bank of (16) physical channels (0-3) */
        }

        if (dev_info->burst_done)                                   /* done w/ burst - disable the timer to prevent more conversions */
        {
            outb(0x34, base_address + CNTR_MODE_REG);    /* reset counter mode: this stops the timer (& prevents channel shifting bug) */
            dev_info->burst_done = FALSE;                                   /* reset for next burst */
        }
        else        /* either not in burst mode, or not yet done w/ a burst */
        {
            /* reset the timer now (to help prevent irq reentrance) */
            outb(0x34, base_address + CNTR_MODE_REG);   /* rewrite counter config - necessary to prevent channel shifting*/
            /* load data acq timer 0 of 8254 chip w/ LSByte of our period  */
            /* writing this first byte stops the timer */
            outb(dev_info->timer_period & 0xff, base_address + DAQ_CNTR_0_REG);
            /* write high byte of our period into timer. this starts timer again */
            outb((dev_info->timer_period >> 8) & 0xff, base_address + DAQ_CNTR_0_REG);
        }

        /* write whole bank of adc samples to rt-fifo */
        rtf_put(RESULT_FIFO_ID, dev_info->sample_bank, (dev_info->chan_cnt + 2) * sizeof(int)); 

        /* for simplicity, write out the dio word and mux word on every interrupt */
        if (dev_info->multiplexing)                                                   /* we're using 2 LSBs of dig out port */
            outb(dev_info->mux|dev_info->dio_output_val, base_address + DAQ_DIGITAL_OUT); /* write byte out the dig output port */
        else
            outb(dev_info->dio_output_val, base_address + DAQ_DIGITAL_OUT);          /* write our byte out the digital output port */

#if 0
        /* the following for testing only. visible from "dmesg". remove ! */
          DEBUG(0, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d \n",
          (int)dev_info->sample_bank[2],(int)dev_info->sample_bank[3],(int)dev_info->sample_bank[4],(int)dev_info->sample_bank[5],
          (int)dev_info->sample_bank[6],(int)dev_info->sample_bank[7],(int)dev_info->sample_bank[8],(int)dev_info->sample_bank[9],
          (int)dev_info->sample_bank[10],(int)dev_info->sample_bank[11],(int)dev_info->sample_bank[12],(int)dev_info->sample_bank[13],
          (int)dev_info->sample_bank[14],(int)dev_info->sample_bank[15],(int)dev_info->sample_bank[16],(int)dev_info->sample_bank[17]);
#endif

/*?!?*/ /* we reset channel scan here to try to prevent "channel shifting"
         * Set to channel scanning mode: write w/ MSB set, then reset.
         * Enable interrupt generation (by FIFO) when conversion results avail.
         * Set chan scanning to MAX_CHAN_SCAN (16) samples.  external interrupts are disabled. */
        outb(ScanEnable | FifoIntEnable | dev_info->channel_scan, base_address + DAQ_CMD_REG_1);
        outb(FifoIntEnable | dev_info->channel_scan, base_address + DAQ_CMD_REG_1);
        daqcard700_clear_fifo(base_address);

        dev_info->chan_cnt = 0;                                                          /* init for next bank of channels */
    }   /* end if got full bank of samples */

    reentry_stat = 0;                                                          /* flag that we finished interrupt */

#if SCOPEING
    outb(0,LP_PORT);                                               /* send the byte out the port */
#endif
}


/*======================================================================
    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.

    When a CARD_REMOVAL event is received, we immediately set a
    private flag to block future accesses to this device.  All the
    functions that actually access the device should check this flag
    to make sure the card is still present.
======================================================================*/
/*************************************/
static int daqcard700_event(event_t event, int priority, event_callback_args_t * args)
{
    dev_link_t *link = args->client_data;

    /*DEBUG(0, "daqcard700_event(0x%06x)\n", event);*/

    switch (event)
    {
    case CS_EVENT_CARD_REMOVAL:
        link->state &= ~DEV_PRESENT;
        if (link->state & DEV_CONFIG)
        {
            ((daqcard700_dev_t *) link->priv)->start = 0;                      /* flag that device not in use */
            link->release.expires = RUN_AT(HZ / 20);
            add_timer(&link->release);
        }
        break;
    case CS_EVENT_CARD_INSERTION:
        link->state |= DEV_PRESENT;
        daqcard700_config(link);
        /* printk("DAQCARD card insertion has been processed.\n"); */
        break;
    case CS_EVENT_PM_SUSPEND:
        link->state |= DEV_SUSPEND;
        /* Fall through to next case... */
    case CS_EVENT_RESET_PHYSICAL:
        /* Mark the device as stopped, to block IO until later */
        ((daqcard700_dev_t *) link->priv)->start = 0;
        if (link->state & DEV_CONFIG)
            CardServices(ReleaseConfiguration, link->handle);
        break;
    case CS_EVENT_PM_RESUME:
        link->state &= ~DEV_SUSPEND;
        /* Fall through to next case... */
    case CS_EVENT_CARD_RESET:
        if (link->state & DEV_CONFIG)
        {
            CardServices(RequestConfiguration, link->handle, &link->conf);     /* ?!? why */
            if (INSTALL_HANDLER)                                               /* if we're using an interrupt handler */
            {
                daqcard700_reset();
            }
            ((daqcard700_dev_t *) link->priv)->start = 1;                      /* mark device as "started", available */
            break;
        }
    }
    return 0;
}                                                                              /* daqcard700_event */

/*********************************/
int init_module(void)                                                         /* this func invoked by card services on load of module */
{
    int ret;
    servinfo_t serv;
    DEBUG(0, "%s\n", version);
    CardServices(GetCardServicesInfo, &serv);
    if (serv.Revision != CS_RELEASE_CODE)
    {
        printk(KERN_NOTICE "daqcard700: Card Services release "
               "does not match!\n");
        return -1;
    }
    ret = rtf_create(RESULT_FIFO_ID, RESULT_FIFO_LEN);                               /* create (outgoing) ADC results fifo */
    if (ret != RESULT_FIFO_LEN)
    {
        printk("Error creating rtfifo %d\n",RESULT_FIFO_ID);
        if (ret == -ENODEV)
            printk("fifo number is too great.\n");
        else if (ret == -EBUSY)
            printk("fifo already in use.\n");
        else if (ret == -ENOMEM)
            printk("memory could not be allocated for fifo.\n");
        else
            printk("return value: %d\n",ret);
    }
    ret = rtf_create(CTRL_FIFO_ID, CTRL_FIFO_LEN);                                   /* create (incoming) control fifo */
    if (ret != CTRL_FIFO_LEN)
    {
        printk("Error creating rtfifo %d\n",CTRL_FIFO_ID);
        if (ret == -ENODEV)
            printk("fifo number is too great.\n");
        else if (ret == -EBUSY)
            printk("fifo already in use.\n");
        else if (ret == -ENOMEM)
            printk("memory could not be allocated for fifo.\n");
        else
            printk("return value: %d\n",ret);
    }
    ret = rtf_create_handler(CTRL_FIFO_ID, Ctl_fifo_handler);                        /* register our handler - invoked when fifo is written */
    if (ret)
        printk("Error creating handler for rtfifo %d: %d",CTRL_FIFO_ID,ret);
    rtf_create(DIG_RESULT_FIFO_ID, DIG_RESULT_FIFO_LEN);                        /* create (outgoing) digital input result fifo */
    if (ret)
        printk("Error creating rtfifo %d: %d",DIG_RESULT_FIFO_ID,ret);

    printk(KERN_INFO "Result rt-fifo: %d, config rt-fifo: %d, dig-input result rt-fifo: %d\n",
            RESULT_FIFO_ID, CTRL_FIFO_ID,DIG_RESULT_FIFO_ID);

    register_pcmcia_driver(&dev_info, &daqcard700_attach, &daqcard700_detach);
    register_symtab(&daqcard700_symtab);
    return 0;
}

/*******************************/
void cleanup_module(void)
{
    int ret;

    DEBUG(0, "daqcard700: unloading\n");

    unregister_pcmcia_driver(&dev_info);
    while (dev_list != NULL)                                                   /* run through all cards on the device llist */
    {
        if (dev_list->state & DEV_CONFIG)
            daqcard700_release((u_long) dev_list);
        daqcard700_detach(dev_list);
    }

    ret = rtf_destroy(RESULT_FIFO_ID);                                               /* free the real time fifos */
    if (ret != 0)
        printk("Error destroying rtfifo %d: %d\n",RESULT_FIFO_ID,ret);
    ret = rtf_destroy(CTRL_FIFO_ID);
    if (ret != 0)
        printk("Error destroying rtfifo %d: %d\n",CTRL_FIFO_ID,ret);
    ret = rtf_destroy(DIG_RESULT_FIFO_ID);
    if (ret != 0)
        printk("Error destroying rtfifo %d: %d\n",DIG_RESULT_FIFO_ID,ret);
}


/* this is the handler invoked when commands are present in rtfifo 1 */
/* all commands from linux user space to the daqcarad driver should use this mechanism, */
/* although 'cardctl' commands work too */
/* the configuration struct 'ctl_msg_struct' is defined in daqcard700-config.h */
/****************************/
int Ctl_fifo_handler(unsigned int fifo)                                       /* handler invoked in rt task when fifo accessed */
{
    struct ctl_msg_struct ctl_msg;                                            /* format for our own little messages to rt task */
    daqcard700_dev_t *dev_info;                                               /* to access daqcard struct */
    ioaddr_t base_address;                                                    /* for base address of daqcard */
    int ret;
    dev_info = (daqcard700_dev_t *) dev_list->priv;                            /* get addr of our daq700 struct */
    base_address = dev_info->base_address;                                     /* io addr of the card */
    #if VERBOSE_MESSAGING
    printk("dev_info addr: %p\n",dev_info); /* just for testing ?!?*/
    printk("card base_addr: %d\n",dev_info->base_address); /* just for testing ?!?*/
    #endif
    
    do                                                                         /* FIXME: ?!? caution, possible infinite loop */
    {
        ret = rtf_get(CTRL_FIFO_ID, &ctl_msg, sizeof(ctl_msg));                /* read a mssg out of control fifo */
        #if VERBOSE_MESSAGING
        printk("got message: %d bytes\n",ret);  /* just for testing ?!?*/
        #endif

        if (ret == sizeof(ctl_msg))                                            /* do the job only if we have a fresh message */
        {
            switch (ctl_msg.conversion_mode)
            {
            case CONTIN_ACQUISITION:
                dev_info->burst_acq = FALSE;                                   /* allow acquisition to be continuous */
                break;
            case BURST_ACQUISITION:
                dev_info->burst_acq = TRUE;                                    /* do a burst of all samples and disable int handler */
                #if VERBOSE_MESSAGING
                printk("set burst acq\n"); /* just for testing ?!?*/
                #endif
                break;
            }

            switch (ctl_msg.command)
            {                                                                  /* see what parent told us to do */
            case START_ACQUISITION:
                if (INSTALL_HANDLER)                                           /* if we're using an interrupt handler */
                {
                    if (first_start_acq)                                        /* first call since reconfiguring driver */
                    {
                        daqcard700_reset();                                        /* call the registered default card reset func */
                        first_start_acq = FALSE;                                /* don't call a complete reset next time */
                    }
                    else
                    {
/*?!?new*/              if (dev_info->burst_acq)                            /* only need to restart timer if in burst mode*/
                        {
/*?!?*/                     /* we must reset channel scan here to try to prevent "channel shifting" */
                            /* set to channel scanning mode: write w/ MSB set, then reset. */
                            /*enable interrupt generation (by FIFO) when conversion results avail.*/
                            /* set chan scanning to MAX_CHAN_SCAN (16) samples.  external interrupts are disabled. */
                            outb(ScanEnable | FifoIntEnable | dev_info->channel_scan, base_address + DAQ_CMD_REG_1);
                            outb(FifoIntEnable | dev_info->channel_scan, base_address + DAQ_CMD_REG_1);
                            daqcard700_clear_fifo(base_address);
                            dev_info->chan_cnt = 0;                             /* in case we stopped in the middle of a "bank" somehow*/
                            dev_info->mux = 0;                                          /* in case we stopped in the middle of a burst */
                            /* restart the timer now so it starts generating conversions */
                            /* don't need to reset counter mode reg here, it was done in irq handler */
                            /* load data acq timer 0 of 8254 chip w/ LSByte of our period */
                            outb(dev_info->timer_period & 0xff, base_address + DAQ_CNTR_0_REG);
                            /* clear hardware FIFO, put latest conversion into fifo, clear stat reg error bits */
                            daqcard700_clear_fifo(base_address);                                   /*clear adc while timer is stopped */
                            /* write high byte of our period into timer. this starts timer again */
                            outb((dev_info->timer_period >> 8) & 0xff, base_address + DAQ_CNTR_0_REG);
                        }
                     }
                     dev_info->start = 1;                                       /* mark device as "started", available */
                }

                /* sample_request_id = ctl_msg.message_id; */                      /* send back the id we recieved form incoming fifo */
#if 0   /* this should be implemented, but irq handler currently resets timer to dev_info parameters ! */
                else
                    /* use ctl_msg parameters to set up acquisition */
                {
                    dev_info->timer_period = ctl_msg.timer_period;             /* this works on second call to this func ! why ? */

                    reg1_cfg_mask |= FifoIntEnable;                            /* always enable fifo interrupts (?!? for now) */
                    if (ctl_msg.num_chan > 1)                                  /* if scanning/sampling more than one channels */
                    {
                        reg1_cfg_mask |= ScanEnable;                           /* set chan scanning bit */
                        reg1_cfg_mask |= ctl_msg.num_chan - 1;                 /* set num channels to scan. -1 corrects for counting *
                                                                                * nums in API */
                    }
                    /* set to channel scanning mode: write w/ MSB set, then reset. 
                     * enable interrupt generation (by FIFO) when conversion
                     * results avail. set chan scanning to MAX_CHAN_SCAN (16) samples.  external interrupts are disabled. */
                    outb(reg1_cfg_mask, dev_info->base_address + DAQ_CMD_REG_1);
                    reg1_cfg_mask &= ~ScanEnable;                              /* clear scan enable bit  */
                    outb(reg1_cfg_mask, dev_info->base_address + DAQ_CMD_REG_1);
                    outb(0x00, base_address + DAQ_CMD_REG_2);                  /* enable data acquisition. (is this really necessary *
                                                                                * ?!?) */
                    /* set FIFO to generate interrupt when >=1 sample available. CLK1 signal comes from IO connector. set to *
                     * single-ended input mode. analog input range is +/- 10V. */
                    outb(dev_info->input_mode, base_address + DAQ_CMD_REG_3);

                    /* load data acq timer 0 of 8254 chip w/ LSByte of our period ?!? */
                    outb(ctl_msg.timer_period & 0xff, base_address + DAQ_CNTR_0_REG);
                    /* clear hardware FIFO, put latest conversion into fifo, clear stat reg error bits */
                    daqcard700_clear_fifo(base_address);                       /* ?!? why this between timer writes */
                    /* write high byte of our period (zero) into timer ? */
                    outb((ctl_msg.timer_period >> 8) & 0xff, base_address + DAQ_CNTR_0_REG);

                    printk("\nctl_msg.timer_period %d, ctl_msg.num_chan %d, input_mode 0x%02x\n",
                           ctl_msg.timer_period, ctl_msg.num_chan,
                           dev_info->input_mode);
                }
                /* dev_info->start = 1; *//* mark device as "started", available */
#endif
                break;
            case STOP_ACQUISITION:
                outb(0x34, base_address + CNTR_MODE_REG);                /* reset the timer, this stops the counting */
                dev_info->start = 0;                                         /* mark device as "stopped", not available */
                dev_info->chan_cnt = 0;                                      /* in case we stopped in the middle of a "bank" */
                dev_info->mux = 0;                                          /* in case we stopped in the middle of a burst */
                first_start_acq = TRUE;                                /* call a complete reset next time we start */
                break;
            /* FIXME: implement this: "-1" in config fields for no change to parameter */
            case CONFIGURE_ACQUISITION:
                first_start_acq = TRUE;                                /* call a complete reset next time we start */
                /* this new timer value will be used every interrupt */
                if (ctl_msg.timer_period >= MAX_SAMPLE_RATE_USEC)             /* make sure user didn't want samples at too high a rate */
                {
                    dev_info->timer_period = ctl_msg.timer_period;            /* set master acq timer period, start acq uses this */
                    #if VERBOSE_MESSAGING
                    printk("timer period\n"); /* just for testing ?!?*/
                    #endif
                }
                if ((ctl_msg.num_adc_banks <= MAX_BANKS) && (ctl_msg.num_adc_banks > 0))
                {
                    dev_info->num_adc_banks = ctl_msg.num_adc_banks;               /* set number of banks for the daqcard to multiplex */
                    #if VERBOSE_MESSAGING
                    printk("num adc banks\n"); /* just for testing ?!?*/
                    #endif
                }
                else
                    dev_info->num_adc_banks = 1;                                /* default is one bank */

                if (ctl_msg.num_chan <= MAX_CHANNELS)                        /* can only sample 1-16 channels on this card */
                {
                    dev_info->channel_scan = ctl_msg.num_chan - 1;            /* set number of channels for the daqcard to scan (1-16) */
                    #if VERBOSE_MESSAGING
                    printk("num chan\n"); /* just for testing ?!?*/
                    #endif
                }
                if (ctl_msg.volt_range == TEN_VOLT)                     /* 10 volt, 5 volt. 2.5 volt. always single ended, for now */
                {
                    dev_info->input_mode &= CLR_2_LSB_INT32;            /* clear the two lsb's of input mode, "0" is 10V range */
                    #if VERBOSE_MESSAGING
                    printk("volt range\n"); /* just for testing ?!?*/
                    #endif
                }
                else if (ctl_msg.volt_range == FIVE_VOLT)
                {
                    dev_info->input_mode &= CLR_2_LSB_INT32;            /* clear the two lsb's of input mode */
                    dev_info->input_mode |= 0x02;                       /* set msb's for 5 volt mode */
                }
                else if (ctl_msg.volt_range == TWO_VOLT)
                {
                    dev_info->input_mode &= CLR_2_LSB_INT32;                  /* clear the two lsb's of input mode */
                    dev_info->input_mode |= 0x03;                             /* set msb's for 2.5 volt mode */
                }

                dev_info->result_numsys = ctl_msg.result_numsys;              /* Decimal, Binary Two's Complement, Bipolar Offset Binary */
                #if VERBOSE_MESSAGING
                printk("num sys\n"); /* just for testing ?!?*/
                #endif
                if (ctl_msg.error_correcting == TRUE)                         /* if user wants to use calibration constants on results */
                    /* set flag in error struct for appropriate voltage mode */
                    (dev_info->calib[dev_info->input_mode & 0x03]).correcting = TRUE; 
                else if (ctl_msg.error_correcting == FALSE)
                {
                    (dev_info->calib[dev_info->input_mode & 0x03]).correcting = FALSE;                
                    #if VERBOSE_MESSAGING
                    printk("no error correcting\n"); /* just for testing ?!?*/
                    #endif
                }
                if ((ctl_msg.num_dig_in_banks > 0) && (ctl_msg.num_dig_in_banks < 5))
                {
                     dev_info->num_dig_in_banks = ctl_msg.num_dig_in_banks;
                    #if VERBOSE_MESSAGING
                    printk("set num dig in banks\n"); /* just for testing ?!?*/
                    #endif
                }
                else
                    dev_info->num_dig_in_banks = 1;

                dev_info->send_dig_ins = TRUE;                        /* this change forces a new digital sample */
                #if VERBOSE_MESSAGING
                printk("sent dig ins\n"); /* just for testing ?!?*/
                #endif
                if ((dev_info->num_adc_banks > 1) || (dev_info->num_dig_in_banks > 1)) /* if any multi-bank requests came in */
                {
                    dev_info->multiplexing = TRUE;
                    #if VERBOSE_MESSAGING
                    printk("multiplexing\n"); /* just for testing ?!?*/
                    #endif
                }
                else
                    dev_info->multiplexing = FALSE;
                break;

            case DIGITAL_OUTPUT:
                dev_info->dio_output_val = ctl_msg.dio_output_val;  /* store the byte */
                if (dev_info->multiplexing) /*((dev_info->num_adc_banks != 1) || (dev_info->num_dig_in_banks))*/
                    dev_info->dio_output_val &= (~0x03);  /* clear 2 LSBs, it wil be ORed w/ mux */
                break;
            default:
                return -EINVAL;
            }
        }
    }
    while (ret == sizeof(ctl_msg));                                           /* rtf_get returns sizeof block, or ... */

    if (ret != 0)                                                             /* if a negative value returned there was a failure */
        return -EINVAL;

    return 0;                                                                 /* normal return value */
}
