/*======================================================================

    daqcard700.h v1.04 (known internally @JHCS as daq700pcs.h)
    
    9-30-1999

    A daqcard700 PCMCIA real-time client driver.

    The contents of this file are subject to the Mozilla Public
    License Version 1.0 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a copy of
    the License at http://www.mozilla.org/MPL/

    Software distributed under the License is distributed on an "AS
    IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
    implied. See the License for the specific language governing
    rights and limitations under the License.

    Originally written by Oleg Subbotin as "daqcard700.h" (oleg@cs.nmt.edu)
    Rev 1.01 by Steve Rosenbluth, JHCS 9-18-98 (stever@silicon.net)
    Rev 1.02 by Steve, JHCS 3-12-1999
    Rev 1.03 by Steve, JHCS 8-30-1999
    Rev 1.04 by Steve, JHCS 9-30-1999

    Copyright(C) 1998, 1999 Jim Henson's Creature Shop.

======================================================================*/
/*  epson dot-matrix printer condense code */

/* offsets of registers from base address */
#define DAQ_CMD_REG_1	 0x00                                                    /* Command Register 1 w/o */
#define DAQ_CMD_REG_2	 0x07                                                    /* Command Register 2 r/w */
#define DAQ_CMD_REG_3	 0x05                                                    /* Command Register 3 w/o */
#define DAQ_STAT_REG_1	 0x00                                                   /* Status Register 1  r/o */
#define DAQ_STAT2	 0x01                                                        /* Status Register 2  r/o */
#define DAQ_AD_FIFO_REG	 0x02                                                  /* A/D FIFO Register  r/o */
#define DAQ_AD_CLR_REG	 0x01                                                   /* A/D Clear Register w/o */
#define DAQ_CNTR_0_REG	 0x08                                                   /* Counter 0 Data Register r/w */
#define DAQ_COUNTER1	 0x09                                                     /* Counter 1 Data Register r/w */
#define DAQ_COUNTER2	 0x0a                                                     /* Counter 2 Data Register r/w */
#define CNTR_MODE_REG	 0x0b                                                     /* Counter Mode Register   w/o */
#define TIMR_CLR_REG_ADDR 0x06                                                 /* Timer Interrupt Clear Register w/o */
#define DAQ_DIGITAL_OUT	 0x04                                                  /* Digital Output Register */
#define DAQ_DIGITAL_IN	 0x05                                                   /* Digital Input Register  */
#define INVERT_MSB_12    0x800                                               /* XOR mask for toggling MSB of 12 bit value */

/* CIS configuration information codes */
#define CISTPL_ADC_CALIBRATION		0x81
#define DAQ_ADC_GAIN_CODE_10V_S		0x00
#define DAQ_ADC_GAIN_CODE_5V_S		0x02
#define DAQ_ADC_GAIN_CODE_2_5V_S	0x03
#define DAQ_ADC_GAIN_CODE_10V_D		0x04
#define DAQ_ADC_GAIN_CODE_5V_D		0x06
#define DAQ_ADC_GAIN_CODE_2_5V_D	0x07
#define DAQ_ADC_GAIN_CODE_DEFAULT	0xff

/* Command Register 1 
 * Bitmap:
 * | SCANEN | CNTINTEN | EXTINTEN | FIFOINTEN | MA3 | MA2 | MA1 | MA0 |
 *
 * SCANEN - Scan Enable
 * CNTINTEN - Counter Interrupt Enable 
 * EXTINTEN - External Interrupt Enable
 * FIFOINTEN - FIFO Interrupt Enable
 * MA3..MA0 - Multiplexer Address
 */

/* Command Register 2 
 * Bitmap:
 * | 0 | 0 | 0 | 0 | 0 | 0 | DISABDAQ | X |
 * DISABDAQ - Disable Data Acquisition
 */

/* Command Register 3
 * Bitmap:
 * | 0 | 0 | FIFOHFINT | 0 | CLK1SRC | DIFF | ARNG1 | ARNG0 |
 * 
 * FIFOHFINT - FIFO Half-Full Interrupt
 * CLK1SRC - Clock 1 Source
 * DIFF - Differential Enable
 * ARNG1..ARNG0 - Analog Input Voltage Range
 */


/* Command Register 1; commands */
enum cmd_reg1
{
    ScanEnable = 0x80,
    CounterIntEnable = 0x40,
    ExternalIntEnable = 0x20,
    FifoIntEnable = 0x10
};

/* Command Register 2 */
#define DisableDataAcquisition 0x02

/* Command Register 3 */
#define	FifoHalfFullInt	0x20
#define ClockSource	0x08
#define DiffEnable	0x04
#define InputVoltageRange10  0x00
#define InputVoltageRange5   0x02
#define InputVoltageRange2_5 0x03

/* Status Register 1 */
enum stat_reg1
{
    FifoHalfFull = 0x40,
    ConversionInProgress = 0x10,
    ExtIntStatus = 0x08,
    CounterIntStatus = 0x04,
    DataError = 0x02,
    DATA_AVAIL = 0x01                                                         /* AND mask to check reg for conversion data */
};

/* Status Register 2 */
#define Overflow	0x02
#define Overrun		0x01

/* Additional structures and defines */

#define DAQ_ERR_OVERFLOW 0x01
#define DAQ_ERR_OVERRUN  0x02
#define DAQ_MAGIC 0x4dac4dac
#ifndef TRUE
    #define TRUE 1
#endif
#ifndef FALSE
    #define FALSE 0
#endif

/****************typedefs*****************/
typedef struct daq_error                                                /* the device struct has one of these for each voltage mode in calib[] */
{
    char valid;                                                         /* if we have a constant to use to correct for card error  */
    char correcting;                                                     /* if user wants to use calibration constants on results */
    int offset_error;
    int gain_error;
} daq_error;

typedef struct
{
    dev_node_t node;
    int irq;
    ioaddr_t base_address;
    daq_error calib[9];                                                       /* calibration structs for each range hold error constants & flags */
    /* 2 - 5V (s); 3 - 2.5V (s);    */
    /* 4 - 10 V (d); 6 - 5 V (d);   */
    /* 7 - 2.5V (d); 8 - default ?!? */
    int timer_period;
    int channel_scan;
    int input_mode;                                                           /* bitmask to DAQ cmd reg 3, irq's, voltage */
    char start;
    int num_adc_banks;  /* number of banks of (16) adc channels the driver should multiplex */
    char burst_acq; /* if nonzero: do a burst of all samples and disable int handler, else continuous acq */    
    int result_numsys;  /* Decimal, Binary Two's Complement, Bipolar Offset Binary */
    int num_dig_in_banks;                                                     /* how many banks of digital input bits: 1 to 4 bytes */
    unsigned char dio_output_val; /* 8 bit val to be written to the digital output port. Only 5 MSB's will get written, 3 LSBs reserved */
    unsigned char old_digital_input_word[4];                                      /* stores input bytes for future comparison */
    char multiplexing;                                                        /* whether using digital out bits 0,1 for external MUX */
    char send_dig_ins;                                                      /* flag for irq handler that inputs have changed */
    int chan_cnt;                                    /* count of individual adc channel conversions */
    int mux;                                               /* current bank of mux on which we get conversions */
    int sample_bank[18];                    /* conversion value storage, init first element to magic ID */
    unsigned char digital_input_word[4];              /* dig input port results. */
    /*char reenable_irq;*/                                                  /* flag for irq handler */
/*    unsigned sample_request_id;*/                                           /* is id of the current samples sent down results fifo */
    char burst_done;                                                  /* flag for irq handler */
} daqcard700_dev_t;                                                           /*  struct of private vars for one instance of device */

extern void daqcard700_reset(void);
extern int daqcard700_write_register(unsigned char, unsigned char);
extern unsigned char daqcard700_read_register(unsigned char);
extern int daqcard700_read_fifo(int *, int, int);
extern void daqcard700_stop_timer(void);
